/*----------------------------------------------------------------
 * channels.c - channel handling routines for libnet
 *----------------------------------------------------------------
 *  libnet is (c) Copyright Chad Catlett and George Foot 1997-1998
 *
 *  Please look in `docs' for details, documentation and
 *  distribution conditions.
 */

#include <stdlib.h>
#include <string.h>

#include "libnet.h"
#include "channels.h"
#include "drivers.h"


/* Maybe this could be static?  Hrm... */
struct __channel_list_t *__net_openchannels;


/* __net_channels_init:  (internal)
 *  Performs internal initialisation.
 */
void __net_channels_init() {
	__net_openchannels = (struct __channel_list_t *) malloc (sizeof(struct __channel_list_t));
	__net_openchannels->next = NULL;
	__net_openchannels->chan = NULL; 
}


/* __net_channels_exit:  (internal)
 *  Internal shutdown routine.
 */
void __net_channels_exit() {
	while (__net_openchannels->next) net_closechannel(__net_openchannels->next->chan);
	__net_openchannels = NULL;
}


/* prepare_new_channel:  (internal)
 *	A little helper function to create a new channel.
 */
static struct __channel_list_t *prepare_new_channel (int type) {
	struct __channel_list_t *ptr;

	ptr = (struct __channel_list_t *) malloc (sizeof(struct __channel_list_t));
	if (!ptr) return NULL;

	ptr->chan = (NET_CHANNEL *) malloc (sizeof (NET_CHANNEL));
	if (!ptr->chan) {
		free(ptr);
		return NULL;
	}

	ptr->chan->type = type;
	ptr->chan->driver = __net_get_driver (type);
	ptr->chan->target_addr = NULL;
	ptr->chan->return_addr = NULL;
	ptr->chan->data = NULL;
	ptr->next = __net_openchannels->next;
	__net_openchannels->next = ptr;

	return ptr;
}


/* net_openchannel:
 *  Opens a communications channel over the specified network
 *  type `type'. It returns a pointer to the NET_CHANNEL struct
 *  created, or NULL on error.
 */
NET_CHANNEL *net_openchannel (int type) {
	struct __channel_list_t *ptr;

	if (!(ptr = prepare_new_channel (type))) return NULL;

	if (ptr->chan->driver->init_channel (ptr->chan)) {   /* driver failure */
		__net_openchannels->next = ptr->next;
		free (ptr->chan);
		free (ptr);
		return NULL;
	}

	return ptr->chan;
}


/* net_openinputchannel:
 *	Opens a channel for input over the specified network type 
 *	`type'.  The settings for the channel are up to the driver;
 *	the Winsock driver for example uses port 24785 if possible
 *	unless overridden by a config file.  This function returns a
 *	pointer to the NET_CHANNEL struct created, or NULL on error.
 */
NET_CHANNEL *net_openinputchannel (int type) {
	struct __channel_list_t *ptr;

	if (!(ptr = prepare_new_channel (type))) return NULL;

	if (ptr->chan->driver->init_inputchannel (ptr->chan)) {   /* driver failure */
		free (ptr->chan);
		free (ptr);
		return NULL;
	}

	return ptr->chan;
}


/* net_closechannel:
 *  Closes a previously opened channel. Returns 0 on success.
 */
int net_closechannel(NET_CHANNEL *channel) {
	struct __channel_list_t *ptr,*optr;

	for (ptr=__net_openchannels; ptr->next && (ptr->next->chan!=channel); ptr=ptr->next);
	if (!ptr->next) return 1;
	optr = ptr->next;
	ptr->next = optr->next;
	free(optr);

	channel->driver->destroy_channel (channel);
	if (channel->target_addr) free (channel->target_addr);
	return 0;
}


/* net_assigntarget:
 *  Assigns a target (`target') to the given channel. The format
 *  of `target' depends upon the specific network type; e.g. for
 *  UDP, it could be a null-terminated string: "127.0.0.1:12345".
 *  Returns zero on success, non-zero on error. `Success' doesn't
 *  guarrantee that the target is reachable.
 */
int net_assigntarget(NET_CHANNEL *channel, char *target) {
	if (channel->target_addr) free(channel->target_addr);
	channel->target_addr = strdup (target);
	return channel->driver->update_target (channel);
}


/* net_getlocaladdress:
 *  Returns the local address of the given channel -- this is
 *  probably the address other computers would use to send into
 *  this channel, but may not be for some network drivers.
 */
char *net_getlocaladdress(NET_CHANNEL *channel) {
	if (channel) return channel->return_addr; else return NULL;
}


/* net_send:
 *  Sends data down a channel, returning zero to indicate
 *  success or non-zero if an error occurs.
 */
int net_send(NET_CHANNEL *channel,void *buffer,int size) {
	return channel->driver->send (channel, buffer, size);
}


/* net_receive:
 *  Receives data from a channel, maximum `maxsize' bytes,
 *  storing the data in `buffer' and pointing `from' to a
 *  string containing the sender's address. Returns the number 
 *  of bytes received. 0 is a valid return type; there was no 
 *  data to read. -1 indicates that an error occured.
 */
int net_receive(NET_CHANNEL *channel,void *buffer,int maxsize,char *from) {
	return channel->driver->recv (channel, buffer, maxsize, from);
}


/* net_query:
 *  Returns nonzero if data is ready for reading from the
 *  given channel.
 */
int net_query(NET_CHANNEL *channel) {
	return channel->driver->query (channel);
}
