/*----------------------------------------------------------------
 * core.c - core routines for libnet
 *----------------------------------------------------------------
 *  libnet is (c) Copyright Chad Catlett and George Foot 1997-1998
 *
 *  Please look in `docs' for details, documentation and
 *  distribution conditions.
 */

#include <stdlib.h>

#include "libnet.h"
#include "channels.h"
#include "drivers.h"


extern NET_DRIVER net_driver_nonet, net_driver_wsock, net_driver_sockets;


typedef struct NET_DRIVER_LIST_ENTRY {
	int type;
	NET_DRIVER *driver;
} NET_DRIVER_LIST_ENTRY;


NET_DRIVER_LIST_ENTRY net_driver_list[] = {
	{       NET_DRIVER_NONET,       &net_driver_nonet       },
	{       NET_DRIVER_WSOCK,       &net_driver_wsock       },
	{       NET_DRIVER_SOCKETS,     &net_driver_sockets     },
	{       0,                      NULL                    }
};


static int detected_drivers = -1;
static int initialised_drivers = 0;
static int initialised = 0;
static int done_atexit = 0;


/* __net_get_driver:  (internal)
 *  Returns a pointer to the VFT for the given driver. NULL if it
 *  doesn't exist.
 */
NET_DRIVER *__net_get_driver (int type) {
	NET_DRIVER_LIST_ENTRY *drv;
	for (drv=net_driver_list; (drv->driver) && (drv->type!=type); drv++);
	return drv->driver;
}


/* exitfunc:  (internal)
 *  Calls net_shutdown. net_init installs this as an atexit
 *  function.
 */
static void exitfunc(void) { net_shutdown(); }


/* net_init:
 *  Initialises the libnet library.
 */
int net_init() {
	if (!initialised) {
		initialised = 1;
		detected_drivers = -1;
		initialised_drivers = 0;

		__net_channels_init();

		if (!done_atexit) {
			atexit(exitfunc);
			done_atexit = 1;
		} 
	}
	return 0;
}


/* net_detectdrivers:
 *  Attempts to detect each given driver; returns a list of
 *  which of the drivers were detected.
 */
int net_detectdrivers(int which) {
	int i,detected;
	NET_DRIVER *driver;

	which &= NET_DRIVER_ALL;
	if (detected_drivers==-1) detected_drivers = 0;

	for (i=1; i & NET_DRIVER_ALL; i<<=1)
		if (i & which) {
			driver = __net_get_driver (i);
			detected = driver && (driver->detect());
			detected_drivers &= ~i;
			detected_drivers |= (detected?i:0);
		}

	return detected_drivers & which;
}


/* net_initdrivers:
 *  Initialises the given drivers. Returns the full list of
 *  successfully initialised drivers.
 */
int net_initdrivers(int which) {
	int i;
	NET_DRIVER *driver;

	if (detected_drivers == -1) net_detectdrivers(which);

	which &= detected_drivers;
	which -= which & initialised_drivers;

	for (i=1; i & NET_DRIVER_ALL; i<<=1)
		if (i & which) {
			driver = __net_get_driver (i);
			if (driver && (!driver->init())) initialised_drivers |= i;
		}

	return initialised_drivers;
}


/* exit_all_drivers:  (internal)
 *  Shuts down all network drivers.
 */
static void exit_all_drivers() {
	int i;
	NET_DRIVER *driver;

	for (i=1; i & NET_DRIVER_ALL; i<<=1)
		if ((i & initialised_drivers) && ((driver = __net_get_driver(i)))) {
			driver->exit();
			initialised_drivers &= ~i;
		}
}


/* net_shutdown:
 *  Shuts everything down nicely, returning 0 on success.
 */
int net_shutdown() {
	if (initialised) {
		__net_channels_exit();
		exit_all_drivers();
		initialised = 0;
		return 0;
	} else return 1;
}


/* bitcount:  (internal)
 *  Utility function, counts the set bits in its argument.
 */
static int bitcount (int x) {
	int c=0;
	while (x) {
		c+=x&1;
		x>>=1;
	}
	return c;
}


/* net_getdrivernames:
 *  Returns a list of driver names; the list contains one entry for 
 *  each driver queried in `which', plus the `nonet' driver and a
 *  terminating entry which has a NULL driver name. The list should 
 *  be freed by the caller.
 */
NET_DRIVERNAME *net_getdrivernames (int which) {
	NET_DRIVERNAME *retval;
	int num,i,a;

	which &= NET_DRIVER_ALL; 
	num = bitcount (which);
	retval = (NET_DRIVERNAME *) malloc (sizeof (NET_DRIVERNAME) * (num + 2));

	a = 0;
	for (i=1; i & NET_DRIVER_ALL; i<<=1)
		if (i & which) {
			retval[a].num = i;
			retval[a].name = __net_get_driver(i)->name;
			a++;
		}

	retval[a].num = 0;
	retval[a].name = __net_get_driver(0)->name;

	retval[a+1].num = -1;
	retval[a+1].name = NULL;

	return retval;
}


