/*----------------------------------------------------------------
 * socks.c - sockets driver for libnet
 *----------------------------------------------------------------
 *  libnet is (c) Copyright Chad Catlett and George Foot 1997-1998
 *
 *  Please look in `docs' for details, documentation and
 *  distribution conditions.
 */

/* If this is not a DJGPP compile (i.e. it's Unixy), we use all this;
 * otherwise we just put in a dummy driver.  For now, at least.
 */
#ifndef DJGPP

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <unistd.h>

/*------------------------- Compatibility junk */

#define SOCKET int
#define INVALID_SOCKET ((SOCKET)-1)
#define SOCKET_ERROR (-1)
#define ioctlsocket ioctl
#define closesocket close

static int stricmp (char *a, char *b) {
  int i;
  for (i = 0; a[i]; i++) {
    if (a[i] < b[i]) return -1;
    if (a[i] > b[i]) return 1;
  }
  return b[i] ? -1 : 0;
}

/*------------------------------------------------ libnet interface */

#include "libnet.h"
#include "config.h"

static char desc[80] = "Internet via Sockets";


static int default_port = 24785;	/* default port */
static int forceport = 0;			/* should we reuse non-vacant ports? */
static int ip_address = 0;			/* IP address */


struct channel_data_t {
	SOCKET sock;
	struct sockaddr_in target_address;
	int address_length;
};

static int detect() {
	return NET_DETECT_YES;
}

static int init() {
  return 0;
}

static int drv_exit() {
  return 0;
}

static char *write_address (struct sockaddr_in *addr) {
	static char p[32];
	unsigned char *q;
	q = (unsigned char *) &(addr->sin_addr.s_addr);
	sprintf (p,"%d.%d.%d.%d:%d",q[0],q[1],q[2],q[3],ntohs(addr->sin_port));
	return p;
}

static int get_address (char *str) {
	char *s;
	unsigned char a[4] = {0};
	int isnumerical = 1;
	int i = 3;

	s = str;
	while ((*s) && (isnumerical)) {
		if (isdigit (*s))
			a[i] = a[i] * 10 + (*s - '0');
		else if (*s == '.') {
			i--;
			if (i == -1) isnumerical = 0;
		} else {
			isnumerical = 0;
		}
		s++;
	}

	if (isnumerical && i) return INADDR_NONE;
	if (!isnumerical) return INADDR_NONE; /* dns_lookup (str, dns_timeout, a); */
	return *(int *)a;
}

static int parse_address (char *str, int *addr, int *port) {
	unsigned char a[4] = {0};
	int i = 0;
	char *t = strchr (str, ':');
	*port = t ? *t = 0, atoi (t+1) : 0;
	*addr = get_address (str);
	if (t) *t = ':';
	return (*addr == INADDR_NONE);
}

static int do_init_channel (NET_CHANNEL *chan, int port, int forceport) {
	struct channel_data_t *data;
	int addrlength;
	struct sockaddr_in sock_addr;
	int a;

	data = (struct channel_data_t *) malloc (sizeof (struct channel_data_t));
	if (!data) return 1;
	chan->data = data;

	data->sock = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (data->sock==INVALID_SOCKET) { free(data); return 2; }
	a=1;
	if (ioctlsocket (data->sock, FIONBIO, (u_long *)&a)) {
		closesocket(data->sock);
		free(data);
		return 3;
	}

	/* if forceport is set, tell socket not to fail its bind operation if port 
	 * is already in use */
	setsockopt (data->sock, SOL_SOCKET, SO_REUSEADDR, (char *)&forceport, sizeof (forceport));

	sock_addr.sin_family      = AF_INET;
	sock_addr.sin_port        = htons(port);
	sock_addr.sin_addr.s_addr = INADDR_ANY;

	addrlength = sizeof (sock_addr);

	if (bind (data->sock, (struct sockaddr *)&sock_addr, addrlength)) {
		closesocket(data->sock);
		free(data);
		return 4;
	}

	getsockname (data->sock, (struct sockaddr *)&sock_addr, &addrlength);
	if (sock_addr.sin_addr.s_addr == INADDR_ANY) sock_addr.sin_addr.s_addr = htonl(ip_address);
	chan->return_addr = strdup (write_address (&sock_addr));

	return 0;
}

static int init_channel (NET_CHANNEL *chan) {
	return do_init_channel (chan, 0, 0);     /* pick any port, [don't force] */
}

static int init_inputchannel (NET_CHANNEL *chan) {
	return do_init_channel (chan, default_port, forceport);
}

static int update_target (NET_CHANNEL *chan) {
	struct channel_data_t *data = chan->data;
	int addr, port;

	if (parse_address (chan->target_addr,&addr,&port)) return 1;
	if (port == 0) port = default_port;
	data->target_address.sin_family = AF_INET;
	data->target_address.sin_port = htons(port);
	data->target_address.sin_addr.s_addr = htonl(addr);

	data->address_length = sizeof (data->target_address);

	return 0;
}

static int destroy_channel (NET_CHANNEL *chan) {
	closesocket (((struct channel_data_t *)chan->data)->sock);
	free (chan->data);
	free (chan->return_addr);
	return 0;
}

static int drv_send (NET_CHANNEL *chan, void *buf, int size) {
	struct channel_data_t *data = chan->data;
	int x = sendto(data->sock,(char *)buf,size,0,(struct sockaddr *)&data->target_address,data->address_length);
	return (x == SOCKET_ERROR);
}

static int drv_recv (NET_CHANNEL *chan, void *buf, int size, char *from) {
	struct channel_data_t *data = chan->data;
	struct sockaddr_in from_addr;

	int size_read, addrsize = sizeof (from_addr);

	size_read = recvfrom (data->sock, (char *)buf, size, 0, (struct sockaddr *) &from_addr, &addrsize);
	if (size_read >= 0) {
		if (from) strcpy (from, write_address (&from_addr));
		return size_read;
	} else
		return -1;
}

static int query (NET_CHANNEL *chan) {
	struct channel_data_t *data = chan->data;
	int x;
	char ch;

	/* This technique sucks... sometime I'll change it to use ioctl option FIONREAD */
	x = recvfrom(data->sock,&ch,1,MSG_PEEK,0,0);
	if ((x>=0) || ((x==SOCKET_ERROR) && (errno==EMSGSIZE))) return 1;
	assert ((x==SOCKET_ERROR) && (errno==EWOULDBLOCK));
	return 0;
}

static void load_config (FILE *fp) {
	char *option, *value;

	if (__libnet_internal__seek_section (fp, "sockets")) return;

	while (__libnet_internal__get_setting (fp, &option, &value) == 0) {

		if (!stricmp (option, "port")) {

			int x = atoi (value);
			if (x) default_port = x;

		} else if (!stricmp (option, "forceport")) {

			forceport = atoi (value);

		} else if (!stricmp (option, "ip")) {

			char *chp = value;
			int addr = 0, i;

			for (i = 0; i < 4; i++) {
				while ((*chp) && (*chp != '.')) chp++;
				if (*chp) *chp++ = 0;
				addr = (addr << 8) + atoi (value);
				value = chp;
			}

			ip_address = addr;

		} /* else if ... etc */
	}
}


NET_DRIVER net_driver_sockets = {
	"Sockets",
	desc,

	detect,
	init,
	drv_exit,

	init_channel,
	init_inputchannel,
	destroy_channel,

	update_target,
	drv_send,
	drv_recv,
	query,

	load_config
};

#else

/* Using djgpp (DOS) so we'll just add a dummy driver which will never
 * be detected. */

#include <stdio.h>
#include <stdlib.h>

#include "libnet.h"

int detect () { return NET_DETECT_NO; }
void load_config (FILE *fp) { (void)fp; }

NET_DRIVER net_driver_sockets = {
	"Sockets", "Dummy driver",
	detect,
	NULL, NULL,
	NULL, NULL, NULL,
	NULL, NULL, NULL, NULL,
	load_config
};

#endif
