/*----------------------------------------------------------------
 * template.c - template network driver for libnet
 *----------------------------------------------------------------
 *  libnet is (c) Copyright Chad Catlett and George Foot 1997-1998
 *
 *  Please look in `docs' for details, documentation and
 *  distribution conditions.
 */


/* How to make a new driver:
 * ~~~~~~~~~~~~~~~~~~~~~~~~~
 * 1) Copy this file to one for your driver.
 * 2) Fill in the blanks -- replace the dummy function bodies with whatever
 *    your driver needs.  The comments tell you what each function should
 *    do.
 * 3) For testing, edit the makefile, adding your driver's filename 
 *    (without the extension) to the list starting `DRIVERS ='.  If your
 *    file has any special dependencies (other than libnet.h) add an entry
 *    for it further down in the makefile, under `drivers/wsock.o: ...'.
 * 4) Add a #define for your driver in include/libnet.h, using the next free 
 *    power of two (e.g. 1, 2, 4, 8, ...).  Add your number onto 
 *    NET_DRIVER_ALL, and also onto NET_DRIVER_ALL_USEFUL if it's a useful 
 *    driver.
 * 5) Add your driver's NET_DRIVER struct to the `extern NET_DRIVER' list.
 * 6) Add an entry to the net_driver_list[] in `core/core.c'.
 * 7) Test it.
 * 8) Send me your source code and any special details (e.g. makefile 
 *    dependencies).
 *
 * Any questions?  Ask me: george.foot@merton.ox.ac.uk
 */


#include "libnet.h"
#include "config.h"


/* desc:
 *  This is a string describing the driver -- it can be static like this,
 *  or it can be set dynamically by the `detect' function, which can add
 *  information on what it detected.
 */
static char template_desc[80] = "Template network driver";

/* detect:
 *  This function returns one of the following:
 *
 *  - NET_DETECT_YES    (if the network driver can definitely work)
 *  - NET_DETECT_MAYBE  (if the network driver might be able to work)
 *  - NET_DETECT_NO     (if the network driver definitely cannot work)
 */
static int template_detect() { return NET_DETECT_YES; }

/* init:
 *  Initialises the driver.  Return zero if successful, non-zero if not.
 */
static int template_init()   { return 0; }

/* exit:
 *  Shuts down the driver, freeing allocated resources, etc.  All channels
 *  will already have been destroyed by the `destroy_channel' function, so 
 *  normally all this needs to do is tidy up anything `init' left untidy.
 *  Return zero if successful, non-zero otherwise.
 */
static int template_exit()   { return 0; }

static char ret_addr = 0;

/* init_channel:
 *  This function will be called when a channel is created, before any 
 *  data is sent.  It should at least set the channel's return address to
 *  something meaningful or at least to a valid string if the driver can't
 *  supply a return address.  It can allocate memory, pointed to by the
 *  `data' field of the channel struct.  Return zero if successful, non-zero
 *  otherwise.
 */
static int template_init_channel    (NET_CHANNEL *chan) { chan->return_addr = &ret_addr; return 0; }

/* destroy_channel:
 *  Called when a channel is being closed.  This should tidy up anything 
 *  that init_channel or subsequent functions have left messy, for example 
 *  it should free any data block that has been allocated.  Return zero if 
 *  successful, non-zero otherwise.
 */
static int template_destroy_channel (NET_CHANNEL *chan) { return 0; }

/* update_target:
 *  This will be called if the target_addr for the channel changes. 
 *  If necessary, the address could be converted into network format here.
 *  Return zero if any conversion was possible, but don't bother checking
 *  whether or not the target is reachable.
 */
static int template_update_target   (NET_CHANNEL *chan) { return 0; }


/* send:
 *  Send the given data block down the channel.  `size' is in bytes.
 *  Return zero if successful, otherwise return non-zero.  Don't bother
 *  checking that the packet arrived at the destination though.
 */
static int template_send (NET_CHANNEL *chan, void *buf, int size) { return 0; }

/* recv:
 *  Receive a packet into the memory block given.  `size' is the maximum
 *  number of bytes to receive.  The sender's address will be put in `from'.
 *  Return the number of bytes received, or -1 if an error occured.  A zero 
 *  return value means there was no data to receive.
 *
 *  Behaviour if more that `size' bytes are waiting is undecided -- either
 *  return an error, or read `size' bytes and return `size'.  The remainder
 *  of the packet should probably be discarded, in either case.  For now,
 *  it's the user's fault for not making the buffer big enough.
 */
static int template_recv (NET_CHANNEL *chan, void *buf, int size, char *from) { return 0; }

/* query:
 *  Returns non-zero if there is data waiting to be read from the channel.
 */
static int template_query (NET_CHANNEL *chan) { return 0; }


/* init_input_channel:
 *  This function will be called when an input channel is created, before any 
 *  data is received.  It should do everything init_channel would do, but may
 *  take extra parameters... blah blah blah I'll document it later.
 */
static int template_init_input_channel    (NET_CHANNEL *chan) { chan->return_addr = &ret_addr; return 0; }


/* load_config:
 *	This will be called once when the library is initialised, inviting the
 *	driver to load configuration information from the passed text file.
 */
static void template_load_config (FILE *fp) { }


NET_DRIVER net_driver_template = {
	"No network",    /* name -- initialise this staticly */
	template_desc,

	template_detect,
	template_init,
	template_exit,

	template_init_channel,
	template_init_input_channel,
	template_destroy_channel,

	template_update_target,
	template_send,
	template_recv,
	template_query,

	template_load_config
};

